"""
pari_pyx.pyx
"""

include 'cdefs.pxi'

cdef extern from 'pari/pari.h':
    ctypedef long* GEN
    GEN     flisseq(char *t)
    char*   GENtostr(GEN x)
    void    cgiv(GEN x)
    ctypedef int size_t
    void    pari_init(size_t parisize, unsigned long maxprime)
    void    pari_sig_init(...)

    GEN     flisexpr(char *t)
    GEN     gmulsg(long s, GEN y)
    GEN     gadd(GEN x, GEN y)
    GEN     gsub(GEN x, GEN y)
    GEN     gmul(GEN x, GEN y)
    GEN     gdiv(GEN x, GEN y)
    GEN     gneg(GEN x)
    int     gcmp(GEN x, GEN y)
    GEN     gmod(GEN x, GEN y)
    GEN     gpowgs(GEN x, long n)
    GEN     gcopy(GEN x)

    # arith1.c 
    GEN     gisprime(GEN x, long flag)


cdef extern from 'setjmp.h':
    struct __jmp_buf_tag:
        pass
    ctypedef __jmp_buf_tag jmp_buf
    int setjmp (jmp_buf __env)

cdef extern from 'pari/anal.h':
    struct anal_xxx:
        jmp_buf env
    ctypedef anal_xxx gp_data
    extern gp_data* GP_DATA


# Initialize the relevant part of GP for exception trapping.
cdef gp_data __GP_DATA
GP_DATA = &__GP_DATA

cdef int initialized
initialized = 0

def init(long size=100000000, unsigned long maxprime=50000):
    """
    Initialize the PARI system.
    
    INPUT:
        size -- long, the number of bytes given to PARI to work with
                (default: 100000000)
        maxprime -- unsigned long, upper limit on a precomputed prime
                    number table  (default: 50000)
    
    If you do not want prime numbers, put maxprime=2, but be careful
    because many PARI functions require this table.  If you get the
    error message "not enough precomputed primes", increase this
    parameter.
    EXAMPLES:
        >>> pari_init()   
    """
    if initialized:
        raise RuntimeError, "pari_init has already been called."
    else:
        pari_init(size, maxprime)
        initialized = 1
        
############################################################
# conversion between GEN and string types
############################################################

cdef GEN str_to_GEN(char* s) except NULL:
    if setjmp(GP_DATA.env):
        raise RuntimeError, "Unable to coerce %s to a PARI"%s
    return flisseq(s)

cdef object GEN_to_str(GEN g):
    cdef char* c
    c = GENtostr(g)
    s = str(c)
    free(c)
    return s


############################################################
# Extension class that models the PARI GEN type.
############################################################

cdef class gen:
    cdef GEN x

    def __init__(self):
        self.x = <GEN> 0

    def __del__(self):
        # TODO -- does this work?
        if self.x:
            cgiv(self.x)
    
    def __repr__(self):
        return GEN_to_str(self.x)


    ###########################################
    # ARITHMETIC
    ###########################################    
    def __add__(gen self, gen other):
        cdef gen o
        o = gen()
        o.x = gadd(self.x, other.x)
        return o

    def __sub__(gen self, gen other):
        cdef gen o
        o = gen()
        o.x = gsub(self.x, other.x)
        return o

    def __mul__(gen self, gen other):
        cdef gen o
        o = gen()
        o.x = gmul(self.x, other.x)
        return o

    def __div__(gen self, gen other):
        cdef gen o
        o = gen()
        o.x = gdiv(self.x, other.x)
        return o

    def __neg__(gen self):
        cdef gen o
        o = gen()
        o.x = gneg(self.x)
        return o

    def __mod__(gen self, gen other):
        cdef gen o
        o = gen()
        o.x = gmod(self.x, other.x)
        return o

    def __pow__(gen self, long n, m):
        cdef gen o
        o = gen()
        o.x = gpowgs(self.x, n)
        return o

    ###########################################
    # ACCESS
    ###########################################
    def __getitem__(gen self, long n):
        cdef gen o
        o = gen()
        o.x = <GEN> self.x[n]
        return o

    ###########################################
    # ?
    ###########################################
    
    def __cmp__(gen self, gen other):
        return gcmp(self.x, other.x)

    def copy(gen self):
        cdef gen o
        o = gen()
        o.x = gcopy(self.x)
        return o

###########################################
# Create a gen from a GEN object.
# This *steals* a reference to the GEN.
###########################################
cdef gen new_gen(GEN x):
    cdef gen z
    z.x = x
    return z

def pari(s):
    cdef gen p
    t = str(s)
    p = gen()
    p.x = str_to_GEN(t)
    return p

###########################################
# arith1.c
###########################################
def is_prime(gen n, long flag):
    return new_gen(gisprime(n.x, flag))




#init()

