"""
pari_pyx.pyx
"""

import sage.misc.misc

include 'cdefs.pxi'

cdef extern from 'pari/pari.h':
    ctypedef long* GEN
    GEN     flisseq(char *t)
    char*   GENtostr(GEN x)
    void    cgiv(GEN x)
    ctypedef int size_t
    void    pari_init(size_t parisize, unsigned long maxprime)
    void    pari_sig_init(...)

    GEN     flisexpr(char *t)
    GEN     gmulsg(long s, GEN y)
    GEN     gadd(GEN x, GEN y)
    GEN     gsub(GEN x, GEN y)
    GEN     gmul(GEN x, GEN y)
    GEN     gdiv(GEN x, GEN y)
    GEN     gneg(GEN x)
    int     gcmp(GEN x, GEN y)
    GEN     gmod(GEN x, GEN y)
    GEN     gpowgs(GEN x, long n)
    GEN     gcopy(GEN x)

#    pass
    #__jmp_buf __jmpbuf          # Calling environment
    #int __mask_was_saved        # Saved the signal mask?  
    #__sigset_t __saved_mask;	# Saved signal mask.  

cdef extern from 'setjmp.h':
    struct __jmp_buf_tag:
        pass
    ctypedef __jmp_buf_tag jmp_buf
    int setjmp (jmp_buf __env)

cdef extern from 'pari/anal.h':
    struct anal_xxx:
        jmp_buf env
    ctypedef anal_xxx gp_data
    extern gp_data* GP_DATA

# Initialize the relevant part of GP for exception trapping.
cdef gp_data __GP_DATA
GP_DATA = &__GP_DATA

initialized = False

# cdef int gp_exception_handler(long numerr):
#     return 0
#     cdef char *s
#     cdef int recovering
#     s = <char*>global_err_data;
#     if s != <char *>0:
#         return 0
#     if s[0]:
#         recovering = 1
#         fprintferr("\n")
#         outerr(lisseq(s))
#         recovering = 0
#         return 0
    #if numerr == errpile:
    #    var_make_safe()
    #    avma = top
    #return break_loop(numerr)


def init(long size=100000000, unsigned long maxprime=50000):
    """
    Initialize the PARI system.
    
    INPUT:
        size -- long, the number of bytes given to PARI to work with
                (default: 100000000)
        maxprime -- unsigned long, upper limit on a precomputed prime
                    number table  (default: 50000)
    
    If you do not want prime numbers, put maxprime=2, but be careful
    because many PARI functions require this table.  If you get the
    error message "not enough precomputed primes", increase this
    parameter.
    EXAMPLES:
        >>> pari_init()   
    """
    global initialized
    if not initialized:
        pari_init(size, maxprime);
        initialized = True
    else:
        raise RuntimeError, "pari_init has already been called."

def execute(s):
    #sage.misc.misc.verbose("Pari command: "+s)
    if not initialized:
        init()
    # Parse s in PARI.
    s = s + "\n" 
    flisseq(s)
    
def eval(s):
    #sage.misc.misc.verbose("Pari command: "+s)
    if not initialized:
        init()
    cdef GEN x
    cdef char *c
    # Parse s in pari and evaluate result.
    s = s + "\n"
    x = flisseq(s)
    # Get result as a C-string
    c = GENtostr(x)
    # De-callocate memory used by x:
    cgiv(x)
    # Convert c to a Python string.
    s = str(c)
    # Free memory used by c
    free(c)
    return s
    
    
    
    

############
#cdef jmp_buf environment
    
cdef GEN str_to_GEN(char* s):
    cdef int n
    if not GP_DATA:
        print "Not setting jump"
    else:
        n = setjmp(GP_DATA.env)
        print "n = ", n
        if n:
            raise RuntimeError, "Unable to coerce %s to a PARI object"%s
    return flisseq(s)

cdef object GEN_to_str(GEN g):
    cdef char* c
    c = GENtostr(g)
    s = str(c)
    free(c)
    return s

cdef class PariObject:
    cdef GEN x

    def __init__(self):
        self.x = <GEN> 0

    def __del__(self):
        if self.x:
            cgiv(self.x)
    
    def __repr__(self):
        return GEN_to_str(self.x)

    def __add__(PariObject self, PariObject other):
        cdef PariObject o
        o = PariObject()
        o.x = gadd(self.x, other.x)
        return o

    def __sub__(PariObject self, PariObject other):
        cdef PariObject o
        o = PariObject()
        o.x = gsub(self.x, other.x)
        return o

    def __mul__(PariObject self, PariObject other):
        cdef PariObject o
        o = PariObject()
        o.x = gmul(self.x, other.x)
        return o

    def __div__(PariObject self, PariObject other):
        cdef PariObject o
        o = PariObject()
        o.x = gdiv(self.x, other.x)
        return o

    def __neg__(PariObject self):
        cdef PariObject o
        o = PariObject()
        o.x = gneg(self.x)
        return o

    def __cmp__(PariObject self, PariObject other):
        return gcmp(self.x, other.x)

    def __mod__(PariObject self, PariObject other):
        cdef PariObject o
        o = PariObject()
        o.x = gmod(self.x, other.x)
        return o

    def __pow__(PariObject self, long n, m):
        cdef PariObject o
        o = PariObject()
        o.x = gpowgs(self.x, n)
        return o

    def copy(PariObject self):
        cdef PariObject o
        o = PariObject()
        o.x = gcopy(self.x)
        return o
        

def Pari(s):
    cdef PariObject p
    t = str(s)
    p = PariObject()
    p.x = str_to_GEN(t)
    return p


init()
