"""
Access to external tables of modular forms and elliptic curves
"""

#*****************************************************************************
#       MANIN: Modular Forms Software, (c) William Stein, 2004
#
#       Copyright (C) 2004 William Stein <was@math.harvard.edu>
#
#  Distributed under the terms of the GNU General Public License (GPL)
#
#    This code is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#    General Public License for more details.
#
#  The full text of the GPL is available at:
#
#                  http://www.gnu.org/licenses/
#*****************************************************************************

import os

import sage.misc.db as db
import sage.misc.errors as errors
import sage.misc.misc as misc
from sage.rings.arith import factor
import sage.pari.py_pari

PATH = "%s/src/data/cremona/"%os.environ["SAGE_HOME"]

class CremonaCurve:
    """
    Class that defines an elliptic curve in the database.
    """
    def __init__(self, conductor, iso_class, number,
                       a_invariants, rank, torsion, tamagawa):
        self.conductor = conductor
        self.iso_class = iso_class
        self.number = number
        self.a_invariants = a_invariants
        self.rank = rank
        self.torsion = torsion
        self.tamagawa = tamagawa   # {p: (p, cp), ...}

    def __repr__(self):
        s = "%s%s%s: invs=%s, rank=%s, tors=%s, cp=%s"%(\
            self.conductor, self.iso_class, self.number, self.a_invariants,
            self.rank, self.torsion, self.tamagawa)
        try:
            s += ", degphi=%s"%self.modular_degree
        except AttributeError:
            pass
        return s

    def tamagawa_product(self):
        x = 1
        for cp in self.tamagawa.itervalues():
            x *= cp
        return x

    def traces(self,n):
        cmd = ("ellan(ellinit(%s),%s)"%(self.a_invariants,n)).replace("L","")
        return eval(sage.pari.py_pari.eval(cmd))


def curves(conductor):
    """
    Return the list of all curves in the database with given
    conductor, grouped into sublists by isogeny class.

    INPUT:
        conductor - an integer or a list of integers.  If a list,
               returns all curves with conductor in the list.
    """
    if isinstance(conductor, list):
        return [curves(x) for x in conductor]
    file = "%s/%s"%(PATH,conductor)
    if not os.path.exists(file+".gz"):
        return []
    return db.load(file, gzip=True)

def optimal_curves(conductor):
    """
    Return the list of all optimal curves in the database
    with given conductor.

    INPUT:
        conductor - an integer or a list of integers.  If a list,
               returns all curves with conductor in the list.
    """
    if isinstance(conductor, list):
        return [optimal_curves(x) for x in conductor]

    return [X[0] for X in curves(conductor)]
    

def save_curves(curves):
    """
    This function saves a list of curves (grouped by isogeny class) to
    disk.  It extracts the conductor, hence filename, from the first
    curve in the list.  If the list is empty, nothing happens.

    This function is mainly useful when you load a list
    of curves of some conductor, add some data about them,
    and want to save it again for later.  Do not use
    this function on a list of just optimal curves, or you'll
    overwrite the file.

    INPUT:
        curves -- list of *all* curves of a given conductor
    """
    if len(curves) == 0:
        return
    level = curves[0][0].conductor
    db.save(curves, "%s/%s"%(PATH,level), gzip=True)

def curve(label):
    """
    Return the curve with given Cremona label.
    """
    if not isinstance(label, str):
        label = str(label)
    i=0
    while i<len(label) and label[i]>='0' and label[i]<='9':
        i += 1
    j=i+1
    if j>len(label):
        label += "A"
    while j<len(label) and not (label[j]>='0' and label[j]<='9'):
        j += 1
    if j>=len(label):
        label += "1"
    level, iso, num = int(label[:i]), label[i:j], int(label[j:])
    iso = iso.upper()
    iso = (ord(iso) - ord('A')) + 26*(len(iso)-1)
    return curves(level)[iso][num-1]



def import_allcurves(file, start, stop):
    """
    Load data from one of Cremona's allcurves files, from start to stop.
    """
    if not os.path.exists(PATH):
        os.makedirs(PATH)
    level = []
    clss = []
    N_last = 0
    for line in open(file).readlines():
        X = line.split()
        if len(X) < 6:
            continue
        N, iso, num, ainvs, rank, tor = line.split()
        N = int(N)
        if N < start:
            continue
        if N != N_last and N > 0:
            print "Saving level", N_last
            db.save(level, "%s/%s"%(PATH, N_last), gzip=True)
            level = []; clss = []
            level.append(clss)
            N_last = N
        if N > stop:
            return
        if N_last == 0:
            N_last = N
        num = int(num)
        ainvs = eval(ainvs)
        rank = int(rank)
        tor = int(tor)
        P = [p for p, _ in factor(N)]
        cp = {}
        for p in P:
            cmd = "elllocalred(ellinit(%s),%s)"%(ainvs, p)
            cmd = cmd.replace("L","")
            x = eval(sage.pari.py_pari.eval(cmd))
            cp[p] = x[3]
        E = CremonaCurve(conductor = N,
                         iso_class = iso,
                         number = num,
                         a_invariants = ainvs,
                         rank = rank,
                         torsion = tor,
                         tamagawa = cp)

        if num == 1 and len(clss) > 0:
            clss = []
            level.append(clss)
        clss.append(E)
    db.save(level, "%s/%s"%(PATH, N), gzip=True)

def import_degphi(file, start, stop):
    """
    Load data from one of Cremona's degphi files, from start to stop.
    """
    if not os.path.exists(PATH):
        os.makedirs(PATH)

    def emit(conductor, data):
        print "Level", conductor
        C = curves(conductor)
        assert len(C) == len(data)
        for i in xrange(len(C)):
            E = C[i][0]
            if conductor != 990:   # known problem for 990H
                assert E.a_invariants == data[i][3]
            assert E.iso_class == data[i][1]
            assert E.conductor == data[i][0]
            E.modular_degree = data[i][2]
        save_curves(C)
        
    data = []
    N_last = 0
    first = True
    for line in open(file).readlines():
        print line
        X = line.split()
        if len(X) < 5:
            continue
        N, iso, degphi, _, model = line.split()
        assert iso[-1] == "1"
        N = int(N)
        if N < start:
            continue
        if N != N_last and N_last > 0:
            emit(N_last, data)
            data = []; 
            N_last = N
        if N > stop:
            return
        if N_last == 0:
            N_last = N
        iso = iso[:-1]
        degphi = int(degphi)
        model = eval(model)
        data.append([N, iso, degphi, model])
    emit(N, data)



